﻿using HEAL.Attic;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Encodings.SymbolicExpressionTreeEncoding;
using HeuristicLab.Parameters;
using System.Collections.Generic;
using System.Linq;

namespace HeuristicLab.Problems.DataAnalysis.Symbolic {
  [Item("User-defined Multi-objective Evaluator", "A multi-objective evaluator that aggregates the output of its evaluator collection.")]
  [StorableType("779F7C15-63E0-4407-A28D-F4DCB1AA8267")]
  public class SymbolicDataAnalysisMultiObjectiveUserDefinedEvaluator<T> : SymbolicDataAnalysisMultiObjectiveEvaluator<T> where T : class, IDataAnalysisProblemData {
    private const string EvaluatorsParameterName = "Evaluators";
    private const string MaximizationParameterName = "Maximization";

    public IFixedValueParameter<ItemList<ISymbolicDataAnalysisSingleObjectiveEvaluator<T>>> EvaluatorsParameter {
      get { return (IFixedValueParameter<ItemList<ISymbolicDataAnalysisSingleObjectiveEvaluator<T>>>)Parameters[EvaluatorsParameterName]; }
    }

    public ILookupParameter<BoolArray> MaximizationParameter {
      get { return (ILookupParameter<BoolArray>)Parameters[MaximizationParameterName]; }
    }

    public ItemList<ISymbolicDataAnalysisSingleObjectiveEvaluator<T>> Evaluators {
      get { return EvaluatorsParameter.Value; }
    }

    public override IEnumerable<bool> Maximization => Evaluators.Select(x => x.Maximization);

    public SymbolicDataAnalysisMultiObjectiveUserDefinedEvaluator() {
      Parameters.Add(new FixedValueParameter<ItemList<ISymbolicDataAnalysisSingleObjectiveEvaluator<T>>>(EvaluatorsParameterName, new ItemList<ISymbolicDataAnalysisSingleObjectiveEvaluator<T>>()));
      Parameters.Add(new LookupParameter<BoolArray>(MaximizationParameterName));
    }

    public SymbolicDataAnalysisMultiObjectiveUserDefinedEvaluator(SymbolicDataAnalysisMultiObjectiveUserDefinedEvaluator<T> original, Cloner cloner) : base(original, cloner) { }

    [StorableConstructor]
    protected SymbolicDataAnalysisMultiObjectiveUserDefinedEvaluator(StorableConstructorFlag deserializing) : base(deserializing) { }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new SymbolicDataAnalysisMultiObjectiveUserDefinedEvaluator<T>(this, cloner);
    }

    public override double[] Evaluate(IExecutionContext context, ISymbolicExpressionTree tree, T problemData, IEnumerable<int> rows) {
      var evaluators = Evaluators;
      var qualities = new double[evaluators.Count];
      for (int i = 0; i < qualities.Length; ++i) {
        qualities[i] = evaluators[i].Evaluate(context, tree, problemData, rows);
      }
      QualitiesParameter.ActualValue = new DoubleArray(qualities);
      return qualities;
    }

    public override IOperation InstrumentedApply() {
      var solution = SymbolicExpressionTreeParameter.ActualValue;
      var problemData = ProblemDataParameter.ActualValue;

      if (MaximizationParameter.ActualValue.Length != Evaluators.Count) {
        MaximizationParameter.ActualValue = new BoolArray(Maximization.ToArray());
      }

      var qualities = Evaluate(ExecutionContext, solution, problemData, problemData.TrainingIndices);
      QualitiesParameter.ActualValue = new DoubleArray(qualities);
      return base.InstrumentedApply();
    }
  }
}